/*
 *  Sensirion SHT1x and SHT7x temperature/humidity sensors
 */

#include "standard.h"
#include "wx.h"

static struct meas {
	word value;  /* result 16 plus ... */
	byte crc;    /*     ... 8 bits. layout important */
	byte cmd;    /* command 8 bits */
	word mask;   /* valid result bits */
} meas[] = {
	{ 0, 0, 0x03, 0x3FFF }, /* 14-bit temperature */
	{ 0, 0, 0x05, 0x0FFF }, /* 12-bit humidity */
};

#define SOFT_RESET 0x1E

static void
update()
{
	short temp;

	/* VDD +5.0v and (default) 12- and 14-bit values assumed */

	/*
	 *  t_fahrenheit = 0.018 * x - 40
	 *  t_celsius    = 0.010 * x - 40
	 *
	 *  half_c       = 0.020 * x - 80
	 */
	temp = DIV(meas[0].value, 50, NULL) - 80;
	thermometer(temp);

#if 1 /* save lots of rom */

	hygrometer(SHT1x_RH_compensated(meas[1].value, temp >> 1));

#else

	temp >>= 1;    /* half_c to t_celsius */

	/*
	 *  RH_true = (t_celsius - 25) * (0.00008 * x + 0.01)
	 *          - 0.0000028 * x * x
	 *          + 0.0405 * x
	 *          - 4
	 *  as 7.25 = (t_celsius - 25) * (2684 * x + 335544)
	 *          - 94 * x * x
	 *          + 1358954 * x
	 *          - 134217728
	 *
	 *  absurd...
	 */
	word x;
	dword rh, tcomp;

	x = meas[1].value;

	tcomp = EMULx(EMUL(2684, x) + 335544, abs(temp - 25));
	if (temp < 25)
		lnegate(&tcomp);

	rh = EMULx(1358954, x)
	   - EMULx(EMUL(x, x), 94)
	   - 134217728
	   + tcomp;

	hygrometer((*(byte *) &rh) >> 1); /* drop fraction */

#endif /* of rom-savings */
}

static bool
unload(struct meas *p)
{
	byte crc;

	if (! SHT1x_result(p))
		return (FALSE);

	if (p->value & ~p->mask)   /* gross error */
		return (FALSE);

	if (p->value == 0)         /* very unlikely */
		return (FALSE);

	crc = crc_type_1(sizeof(p->value), 0, 0x8C, &p->value);

#if DEBUG
	if (debug && p->crc != crc) {
		print_string("crc received 0x");
		print_word(p->crc);
		print_string(", calculated 0x");
		print_word(crc);
		eol();
	}
#endif
	return (p->crc == crc);
}

/* called once every WX_INTERVAL seconds */

void
sht1x_round()
{
	enum { BAD, RESET, OK } state;
	static toggle;

	switch (state) {
	default:
		SHT1x_sync();
		if (SHT1x_command(SOFT_RESET)) {
			toggle = 0;
			state = RESET;
		}
		return; /* come back later */

	case OK:
		if (! unload(&meas[toggle]))
			break;

		if ((toggle ^= 1) == 0) /* have both ? */
			update();

	case RESET:
		state = OK;

		if (SHT1x_command(meas[toggle].cmd))
			return;
	}
	state = BAD; /* unload or command failed */
}

